from typing import Tuple, Optional
import uuid
import random
import logging

from pymilvus import (
    DataType,
    Collection,
    FieldSchema,
    MilvusException,
    CollectionSchema,
    utility,
    connections,
)
from langchain.schema import Document
from langchain_community.embeddings import OpenAIEmbeddings
from langchain_community.vectorstores import Milvus

from .constants import MAX_AGENT_DESC_LENGTH

logger = logging.getLogger(__name__)

DOCUMENT_METADATA_COLLECTION_NAME = "DocumentMetadata1"
DOCUMENT_EMBEDDINGS_COLLECTION_NAME = "DocumentEmbeddings1"

METADATA_VECTOR_DIM = 2
METADATA_FIELDS = [
    "id",
    "name",
    "author",
    "title",
    "format",
    "subject",
    "creator",
    "producer",
    "creationDate",
    "modDate",
    "source",
]


def align_metadata(
    metadata: list[dict], isDeleted: Optional[bool] = False
) -> list[list]:
    """

    Ensure that specific metadata fields are present; if not provided, fill with
    "unknown". Also, add a random vector to each metadata item to simulate an
    embedding.

    Args:
        metadata (List[Dict]): List of metadata items

        isDeleted (Optional[bool], optional): Whether the document is deleted.
            Defaults to False.

    Returns:
        List[List]: List of metadata items, with each item being a list of
            metadata fields.
    """
    ret = []
    fields = METADATA_FIELDS.copy()
    fields.pop(0)
    for ix, k in enumerate(fields):
        ret.append([item[k] if k in item else "unknown" for item in metadata])

    ret.append(
        [
            [random.random() for _ in range(METADATA_VECTOR_DIM)]
            for _ in range(len(metadata))
        ]
    )
    ret.append([isDeleted for _ in metadata])
    return ret


def align_embeddings(docs: list[Document], meta_id: int) -> list[Document]:
    """
    Ensure that the metadata id is present in each document.

    Args:
        docs (List[Document]): List of documents

        meta_id (int): Metadata id to assign to the documents

    Returns:
        List[Document]: List of documents, with each document having a metadata
            id.
    """
    ret = []
    for doc in docs:
        ret.append(
            Document(
                page_content=doc.page_content,
                metadata={"meta_id": meta_id},
            )
        )
    return ret


def validate_connection_args(connection_args: Optional[dict] = None):
    if connection_args is None:
        return {
            "host": "127.0.0.1",
            "port": "19530",
            "user": "",
            "password": "",
        }

    connection_args["user"] = connection_args.get("user", "")
    connection_args["password"] = connection_args.get("password", "")
    return connection_args


class VectorDatabaseAgentMilvus:
    """
    The VectorDatabaseAgentMilvus class manages vector databases in a connected
    host database. It manages an embedding collection
    `_col_embeddings:langchain.vectorstores.Milvus`, which is the main
    information on the embedded text fragments and the basis for similarity
    search, and a metadata collection `_col_metadata:pymilvus.Collection`, which
    stores the metadata of the embedded text fragments. A typical workflow
    includes the following operations:

    1. connect to a host using `connect()`
    2. get all documents in the active database using `get_all_documents()`
    3. save a number of fragments, usually from a specific document, using
        `store_embeddings()`
    4. do similarity search among all fragments of the currently active database
        using `similarity_search()`
    5. remove a document from the currently active database using
        `remove_document()`
    """

    def __init__(
        self,
        embedding_func: OpenAIEmbeddings,
        connection_args: Optional[dict] = None,
        embedding_collection_name: Optional[str] = None,
        metadata_collection_name: Optional[str] = None,
    ):
        """
        Args:
            embedding_func OpenAIEmbeddings: Function used to embed the text

            connection_args Optional dict: args to connect Vector Database

            embedding_collection_name Optional str: exposed for test

            metadata_collection_name Optional str: exposed for test
        """
        self._embedding_func = embedding_func
        self._col_embeddings: Optional[Milvus] = None
        self._col_metadata: Optional[Collection] = None
        self._connection_args = validate_connection_args(connection_args)
        self._embedding_name = (
            embedding_collection_name or DOCUMENT_EMBEDDINGS_COLLECTION_NAME
        )
        self._metadata_name = (
            metadata_collection_name or DOCUMENT_METADATA_COLLECTION_NAME
        )

    def connect(self) -> None:
        """
        Connect to a host and read two document collections (the default names
        are `DocumentEmbeddings` and `DocumentMetadata`) in the currently active
        database (default database name is `default`); if those document
        collections don't exist, create the two collections.
        """
        self._connect(**self._connection_args)
        self._init_host()

    def _connect(self, host: str, port: str, user: str, password: str) -> None:
        self.alias = self._create_connection_alias(host, port, user, password)

    def _init_host(self) -> None:
        """
        Initialize host. Will read/create document collection inside currently
        active database.
        """
        self._create_collections()

    def _create_connection_alias(
        self, host: str, port: str, user: str, password: str
    ) -> str:
        """
        Connect to host and create a connection alias for metadata collection
        using a random uuid.

        Args:
            host (str): host ip address
            port (str): host port

        Returns:
            str: connection alias
        """
        alias = uuid.uuid4().hex
        try:
            connections.connect(
                host=host, port=port, user=user, password=password, alias=alias
            )
            logger.debug(f"Created new connection using: {alias}")
            return alias
        except MilvusException as e:
            logger.error(f"Failed to create  new connection using: {alias}")
            raise e

    def _create_collections(self) -> None:
        """
        Create or load the embedding and metadata collections from the currently
        active database.
        """
        embedding_exists = utility.has_collection(
            self._embedding_name, using=self.alias
        )
        meta_exists = utility.has_collection(
            self._metadata_name,
            using=self.alias,
        )

        if embedding_exists:
            self._load_embeddings_collection()
        else:
            self._create_embeddings_collection()

        if meta_exists:
            self._load_metadata_collection()
        else:
            self._create_metadata_collection()

        self._create_metadata_collection_index()
        self._col_metadata.load()

    def _load_embeddings_collection(self) -> None:
        """
        Load embeddings collection from currently active database.
        """
        try:
            self._col_embeddings = Milvus(
                embedding_function=self._embedding_func,
                collection_name=self._embedding_name,
                connection_args=self._connection_args,
            )
        except MilvusException as e:
            logger.error(
                f"Failed to load embeddings collection {self._embedding_name}."
            )
            raise e

    def _create_embeddings_collection(self) -> None:
        """
        Create embedding collection.
        All fields: "meta_id", "vector"
        """
        try:
            self._col_embeddings = Milvus(
                embedding_function=self._embedding_func,
                collection_name=self._embedding_name,
                connection_args=self._connection_args,
            )
        except MilvusException as e:
            logger.error(
                f"Failed to create embeddings collection {self._embedding_name}"
            )
            raise e

    def _load_metadata_collection(self) -> None:
        """
        Load metadata collection from currently active database.
        """
        self._col_metadata = Collection(
            self._metadata_name,
            using=self.alias,
        )
        self._col_metadata.load()

    def _create_metadata_collection(self) -> None:
        """
        Create metadata collection.

        All fields: "id", "name", "author", "title", "format", "subject",
        "creator", "producer", "creationDate", "modDate", "source", "embedding",
        "isDeleted".

        As the vector database requires a vector field, we will create a fake
        vector "embedding". The field "isDeleted" is used to specify if the
        document is deleted.
        """
        MAX_LENGTH = 10000
        doc_id = FieldSchema(
            name="id", dtype=DataType.INT64, is_primary=True, auto_id=True
        )
        doc_name = FieldSchema(
            name="name", dtype=DataType.VARCHAR, max_length=MAX_LENGTH
        )
        doc_author = FieldSchema(
            name="author", dtype=DataType.VARCHAR, max_length=MAX_LENGTH
        )
        doc_title = FieldSchema(
            name="title", dtype=DataType.VARCHAR, max_length=MAX_LENGTH
        )
        doc_format = FieldSchema(
            name="format", dtype=DataType.VARCHAR, max_length=255
        )
        doc_subject = FieldSchema(
            name="subject", dtype=DataType.VARCHAR, max_length=MAX_LENGTH
        )
        doc_creator = FieldSchema(
            name="creator", dtype=DataType.VARCHAR, max_length=MAX_LENGTH
        )
        doc_producer = FieldSchema(
            name="producer", dtype=DataType.VARCHAR, max_length=MAX_LENGTH
        )
        doc_creationDate = FieldSchema(
            name="creationDate", dtype=DataType.VARCHAR, max_length=1024
        )
        doc_modDate = FieldSchema(
            name="modDate", dtype=DataType.VARCHAR, max_length=1024
        )
        doc_source = FieldSchema(
            name="source", dtype=DataType.VARCHAR, max_length=MAX_LENGTH
        )
        embedding = FieldSchema(
            name="embedding",
            dtype=DataType.FLOAT_VECTOR,
            dim=METADATA_VECTOR_DIM,
        )
        isDeleted = FieldSchema(
            name="isDeleted",
            dtype=DataType.BOOL,
        )
        fields = [
            doc_id,
            doc_name,
            doc_author,
            doc_title,
            doc_format,
            doc_subject,
            doc_creator,
            doc_producer,
            doc_creationDate,
            doc_modDate,
            doc_source,
            embedding,
            isDeleted,
        ]
        schema = CollectionSchema(fields=fields)
        try:
            self._col_metadata = Collection(
                name=self._metadata_name, schema=schema, using=self.alias
            )
        except MilvusException as e:
            logger.error(f"Failed to create collection {self._metadata_name}")
            raise e

    def _create_metadata_collection_index(self) -> None:
        """
        Create index for metadata collection in currently active database.
        """
        if (
            not isinstance(self._col_metadata, Collection)
            or len(self._col_metadata.indexes) > 0
        ):
            return

        index_params = {
            "metric_type": "L2",
            "index_type": "HNSW",
            "params": {"M": 8, "efConstruction": 64},
        }

        try:
            self._col_metadata.create_index(
                field_name="embedding",
                index_params=index_params,
                using=self.alias,
            )
        except MilvusException as e:
            logger.error(
                "Failed to create index for meta collection "
                f"{self._metadata_name}."
            )
            raise e

    def _insert_data(self, documents: list[Document]) -> str:
        """
        Insert documents into the currently active database.

        Args:
            documents (List[Documents]): documents array, usually from
                DocumentReader.load_document, DocumentReader.document_from_pdf,
                DocumentReader.document_from_txt

        Returns:
            str: document id
        """
        if len(documents) == 0:
            return None
        metadata = [documents[0].metadata]
        aligned_metadata = align_metadata(metadata)
        try:
            result = self._col_metadata.insert(aligned_metadata)
            meta_id = str(result.primary_keys[0])
            self._col_metadata.flush()
        except MilvusException as e:
            logger.error(f"Failed to insert meta data")
            raise e
        aligned_docs = align_embeddings(documents, meta_id)
        try:
            # As we passed collection_name, documents will be added to existed collection
            self._col_embeddings = Milvus.from_documents(
                embedding=self._embedding_func,
                collection_name=self._embedding_name,
                connection_args=self._connection_args,
                documents=aligned_docs,
            )
        except MilvusException as e:
            logger.error(
                "Failed to insert data to embedding collection "
                f"{self._embedding_name}."
            )
            raise e
        return meta_id

    def store_embeddings(self, documents: list[Document]) -> str:
        """
        Store documents in the currently active database.

        Args:
            documents (List[Documents]): documents array, usually from
                DocumentReader.load_document, DocumentReader.document_from_pdf,
                DocumentReader.document_from_txt

        Returns:
            str: document id
        """
        if len(documents) == 0:
            return
        return self._insert_data(documents)

    def _build_embedding_search_expression(
        self, meta_ids: list[dict]
    ) -> Optional[str]:
        """
        Build search expression for embedding collection. The generated
        expression follows the pattern: "meta_id in [{id1}, {id2}, ...]

        Args:
            meta_ids: the array of metadata id in metadata collection

        Returns:
            str: search expression or None
        """
        if len(meta_ids) == 0:
            return "meta_id in []"
        built_expr = """meta_id in ["""
        for item in meta_ids:
            id = f'"{item["id"]}",'
            built_expr += id
        built_expr = built_expr[:-1]
        built_expr += """]"""
        return built_expr

    def _join_embedding_and_metadata_results(
        self, result_embedding: list[Document], result_meta: list[dict]
    ) -> list[Document]:
        """
        Join the search results of embedding collection and results of metadata.

        Args:
            result_embedding (List[Document]): search result of embedding
                collection

            result_meta (List[Dict]): search result of metadata collection

        Returns:
            List[Document]: combined results like
                [{page_content: str, metadata: {...}}]
        """

        def _find_metadata_by_id(
            metadata: list[dict], id: str
        ) -> Optional[dict]:
            for d in metadata:
                if str(d["id"]) == id:
                    return d
            return None

        joined_docs = []
        for res in result_embedding:
            found = _find_metadata_by_id(result_meta, res.metadata["meta_id"])
            if found is None:  # discard
                logger.error(
                    f"Failed to join meta_id {res.metadata['meta_id']}"
                )
                continue
            joined_docs.append(
                Document(page_content=res.page_content, metadata=found)
            )
        return joined_docs

    @staticmethod
    def _build_meta_col_query_expr_for_all_documents(
        doc_ids: Optional[list[str]] = None,
    ) -> str:
        """
        Build metadata collection query expression to obtain all documents.

        Args:
            doc_ids: the list of document ids (metadata ids), if thie argument is None,
                     that is, the query is to get all undeleted documents in metadata collection.
                     Otherwise, the query is to getr all undeleted documents form provided doc_ids

        Returns:
            query: str
        """
        expr = (
            f"id in {doc_ids} and isDeleted == false"
            if doc_ids is not None
            else "isDeleted == false"
        )
        return expr.replace('"', "").replace("'", "")

    def similarity_search(
        self, query: str, k: int = 3, doc_ids: Optional[list[str]] = None
    ) -> list[Document]:
        """
        Perform similarity search insider the currently active database
        according to the input query.

        This method will:
        1. get all non-deleted meta_id and build to search expression for
            the currently active embedding collection
        2. do similarity search in the embedding collection
        3. combine metadata and embeddings

        Args:
            query (str): query string

            k (int): the number of results to return

            doc_ids (Optional[list[str]]): the list of document ids, do
                similarity search across the specified documents

        Returns:
            List[Document]: search results
        """
        result_metadata = []
        expr = VectorDatabaseAgentMilvus._build_meta_col_query_expr_for_all_documents(
            doc_ids
        )
        result_metadata = self._col_metadata.query(
            expr=expr, output_fields=METADATA_FIELDS
        )
        expr = self._build_embedding_search_expression(result_metadata)
        result_embedding = self._col_embeddings.similarity_search(
            query=query, k=k, expr=expr
        )
        return self._join_embedding_and_metadata_results(
            result_embedding, result_metadata
        )

    def remove_document(
        self, doc_id: str, doc_ids: Optional[list[str]] = None
    ) -> bool:
        """
        Remove the document include meta data and its embeddings.

        Args:
            doc_id (str): the document to be deleted

            doc_ids (Optional[list[str]]): the list of document ids, defines
                documents scope within which remove operation occurs.

        Returns:
            bool: True if the document is deleted, False otherwise
        """
        if not self._col_metadata:
            return False
        if doc_ids is not None and (
            len(doc_ids) == 0 or (len(doc_ids) > 0 and not doc_id in doc_ids)
        ):
            return False
        try:
            expr = f"id in [{doc_id}]"
            res = self._col_metadata.query(
                expr=expr, output_fields=METADATA_FIELDS
            )
            if len(res) == 0:
                return False
            del_res = self._col_metadata.delete(expr)
            self._col_metadata.flush()

            res = self._col_embeddings.col.query(f'meta_id in ["{doc_id}"]')
            if len(res) == 0:
                return True
            ids = [item["pk"] for item in res]
            embedding_expr = f"pk in {ids}"
            del_res = self._col_embeddings.col.delete(expr=embedding_expr)
            self._col_embeddings.col.flush()
            return True
        except MilvusException as e:
            logger.error(e)
            raise e

    def get_all_documents(
        self, doc_ids: Optional[list[str]] = None
    ) -> list[dict]:
        """
        Get all non-deleted documents from the currently active database.

        Args:
            doc_ids (List[str] optional): the list of document ids, defines
                documents scope within which the operation of obtaining all
                documents occurs

        Returns:
            List[Dict]: the metadata of all non-deleted documents in the form
                [{{id}, {author}, {source}, ...}]
        """
        try:
            expr = VectorDatabaseAgentMilvus._build_meta_col_query_expr_for_all_documents(
                doc_ids
            )
            result_metadata = self._col_metadata.query(
                expr=expr, output_fields=METADATA_FIELDS
            )
            return result_metadata
        except MilvusException as e:
            logger.error(e)
            raise e

    def get_description(self, doc_ids: Optional[list[str]] = None):
        def get_name(meta: dict[str, str]):
            name_col = ["title", "name", "subject", "source"]
            for col in name_col:
                if meta[col] is not None and len(meta[col]) > 0:
                    return meta[col]
            return ""

        expr = VectorDatabaseAgentMilvus._build_meta_col_query_expr_for_all_documents(
            doc_ids
        )
        result = self._col_metadata.query(
            expr=expr,
            output_fields=METADATA_FIELDS,
        )
        names = list(map(get_name, result))
        names_set = set(names)
        desc = f"This vector store contains the following articles: {names_set}"
        return desc[:MAX_AGENT_DESC_LENGTH]



import io
import os
import base64
import tempfile  # needed for test
import subprocess

from PIL import Image
import pdf2image


def convert_and_resize_image(image: Image, max_size: int = 1024) -> Image:
    """
    Convert the image to RGB format if needed and resize it to have a maximum
    dimension of max_size.

    Parameters:
        image (PIL.Image): The input image.
        max_size (int): The maximum size for the image's width or height.

    Returns:
        PIL.Image: The converted and resized PIL image.
    """
    if image.mode != "RGB":
        image = image.convert("RGB")
    image.thumbnail((max_size, max_size), Image.LANCZOS)
    return image


def convert_to_png(image: Image) -> bytes:
    """
    Convert a PIL image to PNG format.

    Parameters:
        image (PIL.Image): The input image.

    Returns:
        bytes: The PNG image data.
    """
    with io.BytesIO() as output:
        image.save(output, format="PNG")
        return output.getvalue()


def convert_to_pil_image(file_path: str, dpi: int = 300) -> Image:
    """
    Convert various image formats (PDF, EPS, TIFF, JPG, PNG) to a PIL image.

    Parameters:
        file_path (str): The path to the image file.
        dpi (int): Dots per inch for high-resolution EPS conversion.

    Returns:
        PIL.Image: The converted PIL image.
    """
    file_path = os.path.abspath(file_path)
    file_ext = os.path.splitext(file_path)[1].lower()

    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")

    if file_ext in [".jpg", ".jpeg", ".png", ".tif", ".tiff"]:
        image = Image.open(file_path)
        return convert_and_resize_image(image)
    elif file_ext == ".pdf":
        pages = pdf2image.convert_from_path(file_path, dpi=dpi)
        if pages:
            return convert_and_resize_image(pages[0])
    elif file_ext == ".eps":
        output_path = file_path.replace(".eps", ".png")
        command = [
            "gs",
            "-dNOPAUSE",
            "-dBATCH",
            "-sDEVICE=pngalpha",
            f"-r{dpi}",
            f"-sOutputFile={output_path}",
            file_path,
        ]
        subprocess.run(command, check=True)
        image = Image.open(output_path)
        return convert_and_resize_image(image)
    else:
        raise ValueError(f"Unsupported file format: {file_ext}")


def process_image(path: str, max_size: int) -> str:
    """
    Process an image, converting it to PNG and resizing if necessary, then
    encode to base64.

    Parameters:
        path (str): The path to the image file.
        max_size (int): The maximum size for the image's width or height.

    Returns:
        str: The base64 encoded image data.
    """
    image = convert_to_pil_image(path)
    png_image = convert_to_png(image)
    return base64.b64encode(png_image).decode("utf-8")


def encode_image(image_path):
    """
    Encode an image file to a base64 string, converting formats if necessary.

    Parameters:
        image_path (str): The path to the image file.

    Returns:
        str: The base64 encoded image data.
    """
    supported_formats = (".webp", ".jpg", ".jpeg", ".gif", ".png")
    file_ext = os.path.splitext(image_path)[1].lower()

    if file_ext in supported_formats:
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode("utf-8")
    else:
        return process_image(image_path, max_size=1024)


def encode_image_from_url(url: str) -> str:
    """
    Download an image from a URL, convert to base64, and return the base64
    string.

    Parameters:
        url (str): The URL of the image.

    Returns:
        str: The base64 encoded image data.
    """
    from urllib.request import urlopen
    import tempfile

    # Get the file extension from the content type
    with urlopen(url) as response:
        content_type = response.info().get_content_type()
        extension = content_type.split("/")[-1]
        extension = (
            "jpg" if extension == "jpeg" else extension
        )  # normalize extension

    with (
        urlopen(url) as response,
        tempfile.NamedTemporaryFile(
            suffix=f".{extension}", delete=False
        ) as tmp_file,
    ):
        tmp_file.write(response.read())
        tmp_file_path = tmp_file.name

    base64_string = encode_image(tmp_file_path)
    os.remove(tmp_file_path)

    return base64_string



from abc import ABC, abstractmethod
from typing import Any, List, Union, Literal, Optional
from datetime import datetime
from collections.abc import Callable
import json
import logging

from langsmith import traceable
from langgraph.graph import END, MessageGraph
from langgraph.graph.graph import CompiledGraph
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ToolMessage,
    HumanMessage,
)
from langchain_core.pydantic_v1 import ValidationError

logger = logging.getLogger(__name__)


class ReflexionAgentLogger:
    def __init__(self) -> None:
        self._logs: str = ""

    def log_step_message(
        self,
        step: int,
        node_name: str,
        output: BaseMessage,
    ):
        """
        log step message
        Args:
          step int: step index
          output BaseMessage: step message
        """
        pass

    def log_final_result(self, final_result: dict[str, Any]) -> None:
        """
        log final result
        Args:
          output BaseMessage: last step message
        """
        pass

    def _log_message(
        self,
        msg: str = "",
        level: Optional[Literal["info", "error", "warn"]] = "info",
    ):
        """
        Save log message

        Args:
            msg: the message to be logged

            level: the log level to write
        """
        logger_func = (
            logger.info
            if level == "info"
            else (logger.error if level == "error" else logger.warning)
        )
        logger_func(msg)
        self._logs = (
            self._logs
            + f"[{level}]"
            + f"{datetime.now().isoformat()} - {msg}\n"
        )

    @property
    def logs(self):
        return self._logs


class ResponderWithRetries:
    """
    Raise request to LLM with 3 retries
    """

    def __init__(self, runnable, validator):
        """
        Args:
        runnable: LLM agent
        validator: used to validate response
        """
        self.runnable = runnable
        self.validator = validator

    @traceable
    def respond(self, state: list[BaseMessage]):
        """
        Invoke LLM agent, this function will be called by LangGraph
        Args:
        state List[BaseMessage]: message history
        """
        response = []
        for attempt in range(3):
            try:
                response = self.runnable.invoke({"messages": state})
                self.validator.invoke(response)
                return response
            except ValidationError as e:
                state = state + [HumanMessage(content=repr(e))]
        return response


DRAFT_NODE = "draft"
EXECUTE_TOOL_NODE = "execute_tool"
REVISE_NODE = "revise"
END_NODE = END


class ReflexionAgentResult:
    def __init__(self, answer: str | None, tool_result: List[Any] | None):
        self.answer = answer
        self.tool_result = tool_result


class ReflexionAgent(ABC):
    """
    LLM agent reflexion framework:

    start -> draft -> execute tool -> revise -> evaluation -> end
                        /|\                        |
                         ---------------------------
    """

    RECURSION_LIMIT = 30

    def __init__(
        self,
        conversation_factory: Callable,
        max_steps: Optional[int] = 20,
        agent_logger: Optional[ReflexionAgentLogger] = ReflexionAgentLogger(),
    ):
        """
        Args:
          conversation_factory Callable: the callback to create Conversation
          max_steps int: max steps for reflextion loop
        """
        if max_steps <= 0:
            max_steps = ReflexionAgent.RECURSION_LIMIT
        recursion_limit = ReflexionAgent.RECURSION_LIMIT
        if recursion_limit < max_steps:
            recursion_limit = max_steps
        self.initial_responder = None
        self.revise_responder = None
        self.max_steps = max_steps
        self.recursion_limit = recursion_limit
        self.conversation = conversation_factory()
        self.agent_logger = agent_logger

    def _should_continue(self, state: list[BaseMessage]):
        """
        Determine if we need to continue reflexion
        Args:
          state List[BaseMessage]: message history
        """
        num_steps = ReflexionAgent._get_num_iterations(state)
        if num_steps > self.max_steps:
            return END
        return EXECUTE_TOOL_NODE

    @abstractmethod
    def _tool_function(self, state: list[BaseMessage]) -> ToolMessage:
        """
        tool function, execute tool based on initial draft or revised answer
        Args:
          state List[BaseMessage]: message history
        Returns:
          ToolMessage
        """
        pass

    @abstractmethod
    def _create_initial_responder(
        self, prompt: Optional[str] = None
    ) -> ResponderWithRetries:
        """
        draft responder, draft initial answer
        Args:
          prompt str: prompt for LLM to draft initial answer
        """
        pass

    @abstractmethod
    def _create_revise_responder(
        self, prompt: Optional[str] = None
    ) -> ResponderWithRetries:
        """
        revise responder, revise answer according to tool function result
        Args:
          prompt str: prompt for LLM to draft initial answer
        """
        pass

    @abstractmethod
    def _parse_final_result(
        self, messages: list[BaseMessage]
    ) -> ReflexionAgentResult:
        """
        parse the result of the last step
        Args:
          output BaseMessage: last step message
        Returns:
          ReflexionAgentResult: the parsed reuslt of the last step
        """
        pass

    def get_logs(self):
        return self.agent_logger.logs

    @staticmethod
    def _get_num_iterations(state: list[BaseMessage]):
        """
        Calculate iteration number
        Args:
          state List[BaseMessage]: message history

        Returns:
          int: the iterations number
        """
        i = 0
        for m in state[::-1]:
            if not isinstance(m, (ToolMessage, AIMessage)):
                break
            i += 1
        return i

    @staticmethod
    def _get_user_question(state: list[BaseMessage]):
        """
        get user's question from messages array
        """
        for m in state:
            if not isinstance(m, HumanMessage):
                continue
            return m.content
        return None

    @staticmethod
    def _get_last_tool_result(messages: list[BaseMessage]):
        """
        get result of the last tool node
        """
        for m in messages[::-1]:
            if not isinstance(m, ToolMessage):
                continue
            content = json.loads(m.content)
            return content["result"]
        return None

    def _build_graph(self, prompt: Optional[str] = None):
        """
        Build Langgraph graph for execution of chained LLM processes.

        Args:
          prompt str: prompt for LLM

        Returns:
          CompiledGraph | None: a Langgraph graph or None in case of errors
        """
        try:
            self.initial_responder = self._create_initial_responder(prompt)
            self.revise_responder = self._create_revise_responder(prompt)
            builder = MessageGraph()
            builder.add_node(DRAFT_NODE, self.initial_responder.respond)
            builder.add_node(EXECUTE_TOOL_NODE, self._tool_function)
            builder.add_node(REVISE_NODE, self.revise_responder.respond)
            builder.add_edge(DRAFT_NODE, EXECUTE_TOOL_NODE)
            builder.add_edge(EXECUTE_TOOL_NODE, REVISE_NODE)

            builder.add_conditional_edges(REVISE_NODE, self._should_continue)
            builder.set_entry_point(DRAFT_NODE)
            graph = builder.compile()
            return graph
        except Exception as e:
            logger.error(e)
            return None

    def _execute_graph(
        self,
        graph: Optional[CompiledGraph] = None,
        question: Optional[str] = "",
    ) -> ReflexionAgentResult:
        """
        execute Langgraph graph
        Args:
          graph CompiledGraph: Langgraph graph
          question str: user question

        Returns:
          answer str | None: string answer parsed from Langgraph graph execution
        """
        if graph is None:
            return None
        if len(question) == 0:
            return None

        events = graph.stream(
            [HumanMessage(content=question)],
            {
                "recursion_limit": self.recursion_limit,
            },
        )
        messages = [HumanMessage(content=question)]
        for i, step in enumerate(events):
            if isinstance(step, list):
                node, output = (f"{i}", step[i])
            else:
                node, output = next(iter(step.items()))
            self.agent_logger.log_step_message(i + 1, node, output)
            messages.append(output)

        final_result = self._parse_final_result(messages)
        self.agent_logger.log_final_result(final_result)
        return final_result

    def execute(
        self, question: str, prompt: Optional[str] = None
    ) -> ReflexionAgentResult:
        """
        Execute ReflexionAgent. Wrapper for building a graph and executing it,
        returning the final answer.

        Args:
          question str: user question
          prompt str: user prompt

        Returns:
          answer str | None: If it executes successfully, an answer to the
            question will be returned, otherwise, it returns None
        """
        if len(question) == 0:
            return None
        graph = self._build_graph(prompt)
        return self._execute_graph(graph, question)



from collections.abc import Callable
import json
from typing import Dict, List, Optional

from langchain.schema import Document
import neo4j_utils as nu

from .prompts import BioCypherPromptEngine
from .constants import MAX_AGENT_DESC_LENGTH
from .kg_langgraph_agent import KGQueryReflexionAgent


class DatabaseAgent:
    def __init__(
        self,
        model_name: str,
        connection_args: dict,
        schema_config_or_info_dict: dict,
        conversation_factory: Callable,
        use_reflexion: bool,
    ) -> None:
        """
        Create a DatabaseAgent analogous to the VectorDatabaseAgentMilvus class,
        which can return results from a database using a query engine. Currently
        limited to Neo4j for development.

        Args:
            connection_args (dict): A dictionary of arguments to connect to the
                database. Contains database name, URI, user, and password.

            conversation_factory (Callable): A function to create a conversation
                for creating the KG query.

            use_reflexion (bool): Whether to use the ReflexionAgent to generate
                the query.
        """
        self.conversation_factory = conversation_factory
        self.prompt_engine = BioCypherPromptEngine(
            model_name=model_name,
            schema_config_or_info_dict=schema_config_or_info_dict,
            conversation_factory=conversation_factory,
        )
        self.connection_args = connection_args
        self.driver = None
        self.use_reflexion = use_reflexion

    def connect(self) -> None:
        """
        Connect to the database and authenticate.
        """
        db_name = self.connection_args.get("db_name")
        uri = f"{self.connection_args.get('host')}:{self.connection_args.get('port')}"
        uri = uri if uri.startswith("bolt://") else "bolt://" + uri
        user = self.connection_args.get("user")
        password = self.connection_args.get("password")
        self.driver = nu.Driver(
            db_name=db_name or "neo4j",
            db_uri=uri,
            user=user,
            password=password,
        )

    def is_connected(self) -> bool:
        return not self.driver is None

    def _generate_query(self, query: str):
        if self.use_reflexion:
            agent = KGQueryReflexionAgent(
                self.conversation_factory,
                self.connection_args,
            )
            query_prompt = self.prompt_engine.generate_query_prompt(query)
            agent_result = agent.execute(query, query_prompt)
            tool_result = (
                [agent_result.tool_result]
                if agent_result.tool_result is not None
                else None
            )
            return agent_result.answer, tool_result
        else:
            query = self.prompt_engine.generate_query(query)
            results = self.driver.query(query=query)
            return query, results

    def _build_response(
        self,
        results: List[Dict],
        cypher_query: str,
        results_num: Optional[int] = 3,
    ) -> List[Document]:
        if len(results) == 0:
            return [
                Document(
                    page_content=(
                        "I didn't find any result in knowledge graph, "
                        f"but here is the query I used: {cypher_query}. "
                        "You can ask user to refine the question. "
                        "Note: please ensure to include the query in a code "
                        "block in your response so that the user can refine "
                        "their question effectively."
                    ),
                    metadata={"cypher_query": cypher_query},
                )
            ]

        clipped_results = results[:results_num] if results_num > 0 else results
        results_dump = json.dumps(clipped_results)

        return [
            Document(
                page_content=(
                    "The results retrieved from knowledge graph are: "
                    f"{results_dump}. "
                    f"The query used is: {cypher_query}. "
                    "Note: please ensure to include the query in a code block "
                    "in your response so that the user can refine "
                    "their question effectively."
                ),
                metadata={"cypher_query": cypher_query},
            )
        ]

    def get_query_results(self, query: str, k: int = 3) -> list[Document]:
        """
        Generate a query using the prompt engine and return the results.
        Replicates vector database similarity search API. Results are returned
        as a list of Document objects to align with the vector database agent.

        Args:
            query (str): A query string.

            k (int): The number of results to return.

        Returns:
            List[Document]: A list of Document objects. The page content values
                are the literal dictionaries returned by the query, the metadata
                values are the cypher query used to generate the results, for
                now.
        """
        (cypher_query, tool_result) = self._generate_query(
            query
        )  # self.prompt_engine.generate_query(query)
        # TODO some logic if it fails?
        if tool_result is not None:
            # If _generate_query() already returned tool_result, we won't connect
            # to graph database to query result any more
            results = tool_result
        else:
            results = self.driver.query(query=cypher_query)

        # return first k results
        # returned nodes can have any formatting, and can also be empty or fewer
        # than k
        if results is None or len(results) == 0 or results[0] is None:
            return []
        return self._build_response(
            results=results[0], cypher_query=cypher_query, results_num=k
        )

    def get_description(self):
        result = self.driver.query("MATCH (n:Schema_info) RETURN n LIMIT 1")

        if result[0]:
            schema_info_node = result[0][0]["n"]
            schema_dict_content = schema_info_node["schema_info"][
                :MAX_AGENT_DESC_LENGTH
            ]  # limit to 1000 characters
            return (
                f"the graph database contains the following nodes and edges: \n\n"
                f"{schema_dict_content}"
            )

        # schema_info is not found in database
        nodes_query = "MATCH (n) RETURN DISTINCT labels(n) LIMIT 300"
        node_results = self.driver.query(query=nodes_query)
        edges_query = "MATCH (n) RETURN DISTINCT type(n) LIMIT 300"
        edge_results = self.driver.query(query=edges_query)
        desc = (
            f"The graph database contains the following nodes and edges: \n"
            f"nodes: \n{node_results}"
            f"edges: \n{edge_results}"
        )
        return desc[:MAX_AGENT_DESC_LENGTH]



from typing import Optional
from collections.abc import Callable


class RagAgentModeEnum:
    VectorStore = "vectorstore"
    KG = "kg"
    API_BLAST = "api_blast"
    API_ONCOKB = "api_oncokb"


class RagAgent:
    def __init__(
        self,
        mode: str,
        model_name: Optional[str] = "gpt-3.5-turbo",
        connection_args: Optional[dict] = None,
        n_results: Optional[int] = 3,
        use_prompt: Optional[bool] = False,
        schema_config_or_info_dict: Optional[dict] = None,
        conversation_factory: Optional[Callable] = None,
        embedding_func: Optional[object] = None,
        documentids_workspace: Optional[list[str]] = None,
        agent_desc: Optional[str] = None,
        use_reflexion: Optional[bool] = False,
    ) -> None:
        ######
        ##TO DO
        ######
        # mode: 'api' for the case where the agent is querying an API
        # use_prompt: default TRUE for self.mode == api

        """
        Create a RAG agent that can return results from a database or vector
        store using a query engine.

        Args:
            mode (str): The mode of the agent. Either "kg" or "vectorstore".

            model_name (str): The name of the model to use.

            connection_args (dict): A dictionary of arguments to connect to the
                database or vector store. Contains database name (in case of
                multiple DBs in one DBMS), host, port, user, and password.

            n_results: the number of results to return for method
                generate_response

            use_prompt (bool): Whether to use the prompt for the query. If
                False, will not retrieve any results and return an empty list.

            schema_config_or_info_dict (dict): A dictionary of schema
                information for the database. Required if mode is "kg".

            conversation_factory (Callable): A function used to create a
                conversation for creating the KG query. Required if mode is
                "kg".

            embedding_func (object): An embedding function. Required if mode is
                "vectorstore".

            documentids_workspace (Optional[List[str]], optional): a list of
                document IDs that defines the scope within which similarity
                search occurs. Defaults to None, which means the operations will
                be performed across all documents in the database.

            agent_desc (str): A description of the agent. If not provided, the
                agent will generate a description based on the mode.

            use_reflexion (bool): Whether to use the ReflexionAgent to generate
                the query.
        """
        self.mode = mode
        self.model_name = model_name
        self.use_prompt = use_prompt
        self.n_results = n_results
        self.documentids_workspace = documentids_workspace
        self.last_response = []
        self._agent_desc = agent_desc
        if self.mode == RagAgentModeEnum.KG:
            from .database_agent import DatabaseAgent

            if not connection_args:
                raise ValueError(
                    "Please provide connection args to connect to database."
                )

            if not schema_config_or_info_dict:
                raise ValueError("Please provide a schema config or info dict.")
            self.schema_config_or_info_dict = schema_config_or_info_dict

            self.agent = DatabaseAgent(
                model_name=model_name,
                connection_args=connection_args,
                schema_config_or_info_dict=self.schema_config_or_info_dict,
                conversation_factory=conversation_factory,
                use_reflexion=use_reflexion,
            )

            self.agent.connect()

            self.query_func = self.agent.get_query_results

        elif self.mode == RagAgentModeEnum.VectorStore:
            from .vectorstore_agent import VectorDatabaseAgentMilvus

            if not connection_args:
                raise ValueError(
                    "Please provide connection args to connect to vector store."
                )

            if not embedding_func:
                raise ValueError("Please provide an embedding function.")

            self.agent = VectorDatabaseAgentMilvus(
                embedding_func=embedding_func,
                connection_args=connection_args,
            )

            self.agent.connect()

            self.query_func = self.agent.similarity_search

        elif self.mode == RagAgentModeEnum.API_BLAST:
            from .api_agent.blast import (
                BlastFetcher,
                BlastInterpreter,
                BlastQueryBuilder,
            )
            from .api_agent.api_agent import APIAgent

            self.agent = APIAgent(
                conversation_factory=conversation_factory,
                query_builder=BlastQueryBuilder(),
                fetcher=BlastFetcher(),
                interpreter=BlastInterpreter(),
            )
            self.query_func = self.agent.execute
        elif self.mode == RagAgentModeEnum.API_ONCOKB:
            from .api_agent.oncokb import (
                OncoKBFetcher,
                OncoKBInterpreter,
                OncoKBQueryBuilder,
            )
            from .api_agent.api_agent import APIAgent

            self.agent = APIAgent(
                conversation_factory=conversation_factory,
                query_builder=OncoKBQueryBuilder(),
                fetcher=OncoKBFetcher(),
                interpreter=OncoKBInterpreter(),
            )
            self.query_func = self.agent.execute
        else:
            raise ValueError(
                "Invalid mode. Choose either 'kg', 'vectorstore', 'api_blast', or 'api_oncokb'."
            )

    @property
    def agent_description(self):
        return self._agent_desc

    @agent_description.setter
    def agent_description(self, val: Optional[str] = None):
        self._agent_desc = val

    def generate_responses(self, user_question: str) -> list[tuple]:
        """
        Run the query function according to the mode and return the results in a
        uniform format (list of tuples, where the first element is the text for
        RAG and the second element is the metadata).

        Args:
            user_question (str): The user question.

        Returns:
            results (List[tuple]): A list of tuples containing the results.

        Todo:
            Which metadata are returned?
        """
        self.last_response = []
        if not self.use_prompt:
            return []
        if self.mode == RagAgentModeEnum.KG:
            results = self.query_func(user_question, self.n_results)
            response = [
                (
                    result.page_content,
                    result.metadata,
                )
                for result in results
            ]
        elif self.mode == RagAgentModeEnum.VectorStore:
            results = self.query_func(
                user_question,
                self.n_results,
                doc_ids=self.documentids_workspace,
            )
            response = [
                (
                    result.page_content,
                    result.metadata,
                )
                for result in results
            ]
        elif self.mode in [
            RagAgentModeEnum.API_BLAST,
            RagAgentModeEnum.API_ONCOKB,
        ]:
            final_answer = self.query_func(user_question)
            if final_answer is not None:
                response = [(final_answer, "response")]
            else:
                response = [(final_answer, "error")]

        else:
            raise ValueError(
                "Invalid mode. Choose either 'kg', 'vectorstore', 'api_blast', "
                "or 'api_oncokb'."
            )
        self.last_response = response
        return response

    def get_description(self):
        if self.agent_description is not None:
            return self.agent_description
        if self.mode == RagAgentModeEnum.KG:
            return self.agent.get_description()
        elif self.mode == RagAgentModeEnum.VectorStore:
            return self.agent.get_description(self.documentids_workspace)
        elif self.mode == RagAgentModeEnum.API_BLAST:
            tool_name = "BLAST"
            tool_desc = (
                "The Basic Local Alignment Search Tool (BLAST) "
                "finds regions of local similarity between sequences. "
                "BLAST compares nucleotide or protein sequences to "
                "sequence databases and calculates the statistical "
                "significance of matches."
            )
            return self.agent.get_description(tool_name, tool_desc)
        elif self.mode == RagAgentModeEnum.API_ONCOKB:
            tool_name = "OncoKB"
            tool_desc = (
                "OncoKB is a precision oncology knowledge base "
                "and contains information about the effects "
                "and treatment implications of specific cancer gene alterations."
            )
            return self.agent.get_description(tool_name, tool_desc)
        else:
            raise ValueError(
                "Invalid mode. Choose either 'kg', 'vectorstore', 'api_blast', "
                "or 'api_oncokb'."
            )



from typing import Optional

from transformers import GPT2TokenizerFast
from langchain.schema import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.embeddings import (
    OllamaEmbeddings,
    XinferenceEmbeddings,
)
from langchain_community.vectorstores import Milvus
from langchain_community.document_loaders import TextLoader
from langchain_community.embeddings.openai import OpenAIEmbeddings
from langchain_community.embeddings.azure_openai import AzureOpenAIEmbeddings
import fitz  # this is PyMuPDF (PyPI pymupdf package, not fitz)
import openai

from biochatter.vectorstore_agent import VectorDatabaseAgentMilvus


class DocumentEmbedder:
    def __init__(
        self,
        used: bool = False,
        online: bool = False,
        chunk_size: int = 1000,
        chunk_overlap: int = 0,
        split_by_characters: bool = True,
        separators: Optional[list] = None,
        n_results: int = 3,
        model: Optional[str] = "text-embedding-ada-002",
        vector_db_vendor: Optional[str] = None,
        connection_args: Optional[dict] = None,
        embedding_collection_name: Optional[str] = None,
        metadata_collection_name: Optional[str] = None,
        api_key: Optional[str] = None,
        is_azure: Optional[bool] = False,
        azure_deployment: Optional[str] = None,
        azure_endpoint: Optional[str] = None,
        base_url: Optional[str] = None,
        embeddings: Optional[
            OpenAIEmbeddings | XinferenceEmbeddings | OllamaEmbeddings
        ] = None,
        documentids_workspace: Optional[list[str]] = None,
    ) -> None:
        """
        Class that handles the retrieval-augmented generation (RAG) functionality
        of BioChatter. It splits text into chunks, embeds them, and stores them in
        a vector database. It can then be used to do similarity search on the
        database.

        Args:

            used (bool, optional): whether RAG has been used (frontend setting).
                Defaults to False.

            online (bool, optional): whether we are running the frontend online.
                Defaults to False.

            chunk_size (int, optional): size of chunks to split text into.
                Defaults to 1000.

            chunk_overlap (int, optional): overlap between chunks. Defaults to 0.

            split_by_characters (bool, optional): whether to split by characters
                or tokens. Defaults to True.

            separators (Optional[list], optional): list of separators to use when
                splitting by characters. Defaults to [" ", ",", "\n"].

            n_results (int, optional): number of results to return from
                similarity search. Defaults to 3.

            model (Optional[str], optional): name of model to use for embeddings.
                Defaults to 'text-embedding-ada-002'.

            vector_db_vendor (Optional[str], optional): name of vector database
                to use. Defaults to Milvus.

            connection_args (Optional[dict], optional): arguments to pass to
                vector database connection. Defaults to None.

            api_key (Optional[str], optional): OpenAI API key. Defaults to None.

            base_url (Optional[str], optional): base url of OpenAI API.

            embeddings (Optional[OpenAIEmbeddings | XinferenceEmbeddings],
                optional): Embeddings object to use. Defaults to OpenAI.

            documentids_workspace (Optional[List[str]], optional): a list of document IDs
                that defines the scope within which rag operations (remove, similarity search,
                and get all) occur. Defaults to None, which means the operations will be
                performed across all documents in the database.

            is_azure (Optional[bool], optional): if we are using Azure
            azure_deployment (Optional[str], optional): Azure embeddings model deployment,
                should work with azure_endpoint when is_azure is True
            azure_endpoint (Optional[str], optional): Azure endpoint, should work with
                azure_deployment when is_azure is True

        """
        self.used = used
        self.online = online
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separators = separators or [" ", ",", "\n"]
        self.n_results = n_results
        self.split_by_characters = split_by_characters
        self.model_name = model

        # TODO API Key handling to central config?
        if base_url:
            openai.api_base = base_url

        if embeddings:
            self.embeddings = embeddings
        else:
            if not self.online:
                self.embeddings = (
                    OpenAIEmbeddings(openai_api_key=api_key, model=model)
                    if not is_azure
                    else AzureOpenAIEmbeddings(
                        api_key=api_key,
                        azure_deployment=azure_deployment,
                        azure_endpoint=azure_endpoint,
                        model=model,
                    )
                )
            else:
                self.embeddings = None

        # connection arguments
        self.connection_args = connection_args or {
            "host": "127.0.0.1",
            "port": "19530",
        }
        self.embedding_collection_name = embedding_collection_name
        self.metadata_collection_name = metadata_collection_name
        self.documentids_workspace = documentids_workspace

        # TODO: vector db selection
        self.vector_db_vendor = vector_db_vendor or "milvus"
        # instantiate VectorDatabaseHost
        self.database_host = None
        self._init_database_host()

    def _set_embeddings(self, embeddings):
        print("setting embedder")
        self.embeddings = embeddings

    def _init_database_host(self):
        if self.vector_db_vendor == "milvus":
            self.database_host = VectorDatabaseAgentMilvus(
                embedding_func=self.embeddings,
                connection_args=self.connection_args,
                embedding_collection_name=self.embedding_collection_name,
                metadata_collection_name=self.metadata_collection_name,
            )
        else:
            raise NotImplementedError(self.vector_db_vendor)

    def set_chunk_siue(self, chunk_size: int) -> None:
        self.chunk_size = chunk_size

    def set_chunk_overlap(self, chunk_overlap: int) -> None:
        self.chunk_overlap = chunk_overlap

    def set_separators(self, separators: list) -> None:
        self.separators = separators

    def _characters_splitter(self) -> RecursiveCharacterTextSplitter:
        return RecursiveCharacterTextSplitter(
            chunk_size=self.chunk_size,
            chunk_overlap=self.chunk_overlap,
            separators=self.separators,
        )

    def _tokens_splitter(self) -> RecursiveCharacterTextSplitter:
        DEFAULT_OPENAI_MODEL = "gpt-3.5-turbo"
        HUGGINGFACE_MODELS = ["bigscience/bloom"]
        if self.model_name and self.model_name in HUGGINGFACE_MODELS:
            tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")
            return RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
                tokenizer,
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap,
                separators=self.separators,
            )
        else:
            return RecursiveCharacterTextSplitter.from_tiktoken_encoder(
                encoding_name="",
                model_name=(
                    DEFAULT_OPENAI_MODEL
                    if not self.model_name
                    else self.model_name
                ),
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap,
                separators=self.separators,
            )

    def _text_splitter(self) -> RecursiveCharacterTextSplitter:
        return (
            self._characters_splitter()
            if self.split_by_characters
            else self._tokens_splitter()
        )

    def save_document(self, doc: list[Document]) -> str:
        """
        This function saves document to the vector database
        Args:
            doc List[Document]: document content, read with DocumentReader load_document(),
                or document_from_pdf(), document_from_txt()
        Returns:
            str: document id, which can be used to remove an uploaded document with remove_document()
        """
        splitted = self._split_document(doc)
        return self._store_embeddings(splitted)

    def _split_document(self, document: list[Document]) -> list[Document]:
        text_splitter = self._text_splitter()
        return text_splitter.split_documents(document)

    def _store_embeddings(self, doc: list[Document]) -> str:
        return self.database_host.store_embeddings(documents=doc)

    def connect(self) -> None:
        self.database_host.connect()

    def get_all_documents(self) -> list[dict]:
        return self.database_host.get_all_documents(
            doc_ids=self.documentids_workspace
        )

    def remove_document(self, doc_id: str) -> None:
        return self.database_host.remove_document(
            doc_id, self.documentids_workspace
        )


class XinferenceDocumentEmbedder(DocumentEmbedder):
    def __init__(
        self,
        used: bool = False,
        chunk_size: int = 1000,
        chunk_overlap: int = 0,
        split_by_characters: bool = True,
        separators: Optional[list] = None,
        n_results: int = 3,
        model: Optional[str] = "auto",
        vector_db_vendor: Optional[str] = None,
        connection_args: Optional[dict] = None,
        embedding_collection_name: Optional[str] = None,
        metadata_collection_name: Optional[str] = None,
        api_key: Optional[str] = "none",
        base_url: Optional[str] = None,
        documentids_workspace: Optional[list[str]] = None,
    ):
        """
        Extension of the DocumentEmbedder class that uses Xinference for
        embeddings.

        Args:

            used (bool, optional): whether RAG has been used (frontend setting).

            chunk_size (int, optional): size of chunks to split text into.

            chunk_overlap (int, optional): overlap between chunks.

            split_by_characters (bool, optional): whether to split by characters
            or tokens.

            separators (Optional[list], optional): list of separators to use when
            splitting by characters.

            n_results (int, optional): number of results to return from
            similarity search.

            model (Optional[str], optional): name of model to use for embeddings.
            Can be "auto" to use the first available model.

            vector_db_vendor (Optional[str], optional): name of vector database
            to use.

            connection_args (Optional[dict], optional): arguments to pass to
            vector database connection.

            embedding_collection_name (Optional[str], optional): name of
            collection to store embeddings in.

            metadata_collection_name (Optional[str], optional): name of
            collection to store metadata in.

            api_key (Optional[str], optional): Xinference API key.

            base_url (Optional[str], optional): base url of Xinference API.

            documentids_workspace (Optional[List[str]], optional): a list of document IDs
            that defines the scope within which rag operations (remove, similarity search,
            and get all) occur. Defaults to None, which means the operations will be
            performed across all documents in the database.

        """
        from xinference.client import Client

        self.model_name = model
        self.client = Client(base_url=base_url)
        self.models = {}
        self.load_models()

        if self.model_name is None or self.model_name == "auto":
            self.model_name = self.list_models_by_type("embedding")[0]
        self.model_uid = self.models[self.model_name]["id"]

        super().__init__(
            used=used,
            online=True,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            split_by_characters=split_by_characters,
            separators=separators,
            n_results=n_results,
            model=model,
            vector_db_vendor=vector_db_vendor,
            connection_args=connection_args,
            embedding_collection_name=embedding_collection_name,
            metadata_collection_name=metadata_collection_name,
            api_key=api_key,
            base_url=base_url,
            embeddings=XinferenceEmbeddings(
                server_url=base_url, model_uid=self.model_uid
            ),
            documentids_workspace=documentids_workspace,
        )

    def load_models(self) -> None:
        """
        Get all models that are currently available on the Xinference server and
        write them to `self.models`.
        """
        for id, model in self.client.list_models().items():
            model["id"] = id
            self.models[model["model_name"]] = model

    def list_models_by_type(self, type: str) -> list[str]:
        """
        Return all models of a certain type that are currently available on the
        Xinference server.

        Args:
            type (str): type of model to list (e.g. "embedding", "chat")

        Returns:
            List[str]: list of model names
        """
        names = []
        for name, model in self.models.items():
            if "model_ability" in model:
                if type in model["model_ability"]:
                    names.append(name)
            elif model["model_type"] == type:
                names.append(name)
        return names


class OllamaDocumentEmbedder(DocumentEmbedder):
    def __init__(
        self,
        used: bool = False,
        chunk_size: int = 1000,
        chunk_overlap: int = 0,
        split_by_characters: bool = True,
        separators: Optional[list] = None,
        n_results: int = 3,
        model: Optional[str] = "nomic-embed-text",
        vector_db_vendor: Optional[str] = None,
        connection_args: Optional[dict] = None,
        embedding_collection_name: Optional[str] = None,
        metadata_collection_name: Optional[str] = None,
        api_key: Optional[str] = "none",
        base_url: Optional[str] = None,
        documentids_workspace: Optional[list[str]] = None,
    ):
        """
        Extension of the DocumentEmbedder class that uses Ollama for
        embeddings.

        Args:

            used (bool, optional): whether RAG has been used (frontend setting).

            chunk_size (int, optional): size of chunks to split text into.

            chunk_overlap (int, optional): overlap between chunks.

            split_by_characters (bool, optional): whether to split by characters
            or tokens.

            separators (Optional[list], optional): list of separators to use when
            splitting by characters.

            n_results (int, optional): number of results to return from
            similarity search.

            model (Optional[str], optional): name of model to use for embeddings.
            Can be "auto" to use the first available model.

            vector_db_vendor (Optional[str], optional): name of vector database
            to use.

            connection_args (Optional[dict], optional): arguments to pass to
            vector database connection.

            embedding_collection_name (Optional[str], optional): name of
            collection to store embeddings in.

            metadata_collection_name (Optional[str], optional): name of
            collection to store metadata in.

            api_key (Optional[str], optional): Xinference API key.

            base_url (Optional[str], optional): base url of Xinference API.

            documentids_workspace (Optional[List[str]], optional): a list of document IDs
            that defines the scope within which rag operations (remove, similarity search,
            and get all) occur. Defaults to None, which means the operations will be
            performed across all documents in the database.

        """
        from langchain_community.embeddings import OllamaEmbeddings

        self.model_name = model

        super().__init__(
            used=used,
            online=True,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            split_by_characters=split_by_characters,
            separators=separators,
            n_results=n_results,
            model=model,
            vector_db_vendor=vector_db_vendor,
            connection_args=connection_args,
            embedding_collection_name=embedding_collection_name,
            metadata_collection_name=metadata_collection_name,
            api_key=api_key,
            base_url=base_url,
            embeddings=OllamaEmbeddings(
                base_url=base_url, model=self.model_name
            ),
            documentids_workspace=documentids_workspace,
        )


class DocumentReader:
    def load_document(self, path: str) -> list[Document]:
        """
        Loads a document from a path; accepts txt and pdf files. Txt files are
        loaded as-is, pdf files are converted to text using fitz.

        Args:
            path (str): path to document

        Returns:
            List[Document]: list of documents
        """
        if path.endswith(".txt"):
            loader = TextLoader(path)
            return loader.load()

        elif path.endswith(".pdf"):
            doc = fitz.open(path)
            text = ""
            for page in doc:
                text += page.get_text()

            meta = {k: v for k, v in doc.metadata.items() if v}
            meta.update({"source": path})

            return [
                Document(
                    page_content=text,
                    metadata=meta,
                )
            ]

    def document_from_pdf(self, pdf: bytes) -> list[Document]:
        """
        Receive a byte representation of a pdf file and return a list of Documents
        with metadata.

        Args:
            pdf (bytes): byte representation of pdf file

        Returns:
            List[Document]: list of documents
        """
        doc = fitz.open(stream=pdf, filetype="pdf")
        text = ""
        for page in doc:
            text += page.get_text()

        meta = {k: v for k, v in doc.metadata.items() if v}
        meta.update({"source": "pdf"})

        return [
            Document(
                page_content=text,
                metadata=meta,
            )
        ]

    def document_from_txt(self, txt: bytes) -> list[Document]:
        """
        Receive a byte representation of a txt file and return a list of Documents
        with metadata.

        Args:
            txt (bytes): byte representation of txt file

        Returns:
            List[Document]: list of documents
        """
        meta = {"source": "txt"}
        return [
            Document(
                page_content=txt,
                metadata=meta,
            )
        ]



from typing import Optional
from collections.abc import Callable

from pydantic import BaseModel

from .abc import BaseFetcher, BaseInterpreter, BaseQueryBuilder

### logic

# 1. User asks question req. API to find the answer
# 2. rag_agent receives the question and sends it to the api_agent // api_agent is manually selected
# 3. api_agent writes query for API specific to the question
# 3.1 question + API prompt template + BlastQuery are input into
# langchain.chains.openai_functions.create_structured_output_runnable -> returns a structured output which is our API call object
# 4. API call
# 4.1 API call is made using the structured output from 3.1
# 4.2 API returns a response which is saved
#
# 5. Read response from and uses it to answer question
# 6. answer is returned to rag_agent


## Agent class
class APIAgent:
    def __init__(
        self,
        conversation_factory: Callable,
        query_builder: "BaseQueryBuilder",
        fetcher: "BaseFetcher",
        interpreter: "BaseInterpreter",
    ):
        """

        API agent class to interact with a tool's API for querying and fetching
        results.  The query fields have to be defined in a Pydantic model
        (`BaseModel`) and used (i.e., parameterised by the LLM) in the query
        builder. Specific API agents are defined in submodules of this directory
        (`api_agent`). The agent's logic is implemented in the `execute` method.

        Attributes:
            conversation_factory (Callable): A function used to create a
                BioChatter conversation, providing LLM access.

            query_builder (BaseQueryBuilder): An instance of a child of the
                BaseQueryBuilder class.

            result_fetcher (BaseFetcher): An instance of a child of the
                BaseFetcher class.

            result_interpreter (BaseInterpreter): An instance of a child of the
                BaseInterpreter class.
        """
        self.conversation_factory = conversation_factory
        self.query_builder = query_builder
        self.fetcher = fetcher
        self.interpreter = interpreter
        self.final_answer = None

    def parameterise_query(self, question: str) -> Optional[BaseModel]:
        """
        Use LLM to parameterise a query (a Pydantic model) based on the given
        question using a BioChatter conversation instance.
        """
        try:
            conversation = self.conversation_factory()
            return self.query_builder.parameterise_query(question, conversation)
        except Exception as e:
            print(f"Error generating query: {e}")
            return None

    def fetch_results(self, query_model: str) -> Optional[str]:
        """
        Fetch the results of the query using the individual API's implementation
        (either single-step or submit-retrieve).

        Args:
            query_model: the parameterised query Pydantic model
        """
        try:
            return self.fetcher.fetch_results(query_model, 100)
        except Exception as e:
            print(f"Error fetching results: {e}")
            return None

    def summarise_results(
        self, question: str, response_text: str
    ) -> Optional[str]:
        """
        Summarise the retrieved results to extract the answer to the question.
        """
        try:
            return self.interpreter.summarise_results(
                question=question,
                conversation_factory=self.conversation_factory,
                response_text=response_text,
            )
        except Exception as e:
            print(f"Error extracting answer: {e}")
            return None

    def execute(self, question: str) -> Optional[str]:
        """
        Wrapper that uses class methods to execute the API agent logic. Consists
        of 1) query generation, 2) query submission, 3) results fetching, and
        4) answer extraction. The final answer is stored in the final_answer
        attribute.

        Args:
            question (str): The question to be answered.
        """
        # Generate query
        try:
            query_model = self.parameterise_query(question)
            if not query_model:
                raise ValueError("Failed to generate query.")
        except ValueError as e:
            print(e)

        # Fetch results
        try:
            response_text = self.fetch_results(
                query_model=query_model,
            )
            if not response_text:
                raise ValueError("Failed to fetch results.")
        except ValueError as e:
            print(e)

        # Extract answer from results
        try:
            final_answer = self.summarise_results(question, response_text)
            if not final_answer:
                raise ValueError("Failed to extract answer from results.")
        except ValueError as e:
            print(e)

        self.final_answer = final_answer
        return final_answer

    def get_description(self, tool_name: str, tool_desc: str):
        return (
            f"This API agent interacts with {tool_name}'s API for querying and "
            f"fetching results. {tool_desc}"
        )



from typing import Optional
from urllib.parse import urlencode
from collections.abc import Callable
import re
import time
import uuid

from pydantic import Field, BaseModel
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.pydantic_v1 import Field, BaseModel
from langchain_core.output_parsers import StrOutputParser
from langchain.chains.openai_functions import create_structured_output_runnable
import requests

from biochatter.llm_connect import Conversation
from .abc import BaseFetcher, BaseInterpreter, BaseQueryBuilder

BLAST_QUERY_PROMPT = """
You are a world class algorithm for creating queries in structured formats. Your task is to use NCBI Web APIs to answer genomic questions.

For questions about DNA sequences (other than genome alignments) you can use BLAST by: "[https://blast.ncbi.nlm.nih.gov/blast/Blast.cgi?CMD={Put|Get}&PROGRAM=blastn&MEGABLAST=on&DATABASE=nt&FORMAT_TYPE={XML|Text}&QUERY={sequence}&HITLIST_SIZE={max_hit_size}]".
BLAST maps a specific DNA {sequence} to DNA sequences among different specices.
Use it if you need information about a DNA sequence, see example questions 1.
You need to first PUT the BLAST request and then GET the results using the RID returned by PUT.

Example Question 1: Which organism does this DNA sequence come from: AGGGGCAGCAAACACCGGGACACACCCATTCGTGCACTAATCAGAAACTTTTTTTTCTCAAATAATTCAAACAATCAAAATTGGTTTTTTCGAGCAAGGTGGGAAATTTTTCGAT
Use BLASTn for such a question -> [https://blast.ncbi.nlm.nih.gov/Blast.cgi??CMD=Put&PROGRAM=blastn&DATABASE=nt&QUERY=AGGGGCAGCAAACACCGGGACACACCCATTCGTGCACTAATCAGAAACTTTTTTTTCTCAAATAATTCAAACAATCAAAATTGGTTTTTTCGAGCAAGGTGGGAAATTTTTCGAT&FORMAT_TYPE=Text&MEGABLAST=on&HITLIST_SIZE=15&email=noah.bruderer%40uib.no]

For questions about protein sequences you can also use BLAST, but this time by: [https://blast.ncbi.nlm.nih.gov/Blast.cgi?CMD=Put&PROGRAM=blastp&DATABASE=nr&FORMAT_TYPE=XML&QUERY=sequence&HITLIST_SIZE=max_hit_size]
Example Question 2: What do you find about protein sequence: MEEPQSDPSV
Use BLASTp for such a question -> https://blast.ncbi.nlm.nih.gov/Blast.cgi?CMD=Put&PROGRAM=blastp&DATABASE=nr&FORMAT_TYPE=XML&QUERY=MEEPQSDPSV&HITLIST_SIZE=10"
"""

BLAST_SUMMARY_PROMPT = """
        You have to answer this question in a clear and concise manner: {question} Be factual!\n\
        If you are asked what organism a specific sequence belongs to, check the 'Hit_def' fields. If you find a synthetic construct or predicted entry, move to the next one and look for an organism name.\n\
        Try to use the hits with the best identity score to answer the question. If it is not possible, move to the next one.\n\
        Be clear, and if organism names are present in ANY of the results, please include them in the answer. Do not make up information and mention how relevant the found information is based on the identity scores.\n\
        Use the same reasoning for any potential BLAST results. If you find information that is manually curated, please use it and state it. You may also state other results, but always include the context.\n\
        Based on the information given here:\n\
        {context}
        """


class BlastQueryParameters(BaseModel):
    """

    BlastQuery is a Pydantic model for the parameters of a BLAST query request,
    used for configuring and sending a request to the NCBI BLAST query API. The
    fields are dynamically configured by the LLM based on the user's question.

    """

    url: Optional[str] = Field(
        default="https://blast.ncbi.nlm.nih.gov/Blast.cgi?",
        description="ALWAYS USE DEFAULT, DO NOT CHANGE",
    )
    cmd: Optional[str] = Field(
        default="Put",
        description="Command to execute, 'Put' for submitting query, 'Get' for retrieving results.",
    )
    program: Optional[str] = Field(
        default="blastn",
        description="BLAST program to use, e.g., 'blastn' for nucleotide-nucleotide BLAST, 'blastp' for protein-protein BLAST.",
    )
    database: Optional[str] = Field(
        default="nt",
        description="Database to search, e.g., 'nt' for nucleotide database, 'nr' for non redundant protein database, pdb the Protein Data Bank database, which is used specifically for protein structures, 'refseq_rna' and 'refseq_genomic': specialized databases for RNA sequences and genomic sequences",
    )
    query: Optional[str] = Field(
        None,
        description="Nucleotide or protein sequence for the BLAST or blat query, make sure to always keep the entire sequence given.",
    )
    format_type: Optional[str] = Field(
        default="Text",
        description="Format of the BLAST results, e.g., 'Text', 'XML'.",
    )
    rid: Optional[str] = Field(
        None, description="Request ID for retrieving BLAST results."
    )
    other_params: Optional[dict] = Field(
        default={"email": "user@example.com"},
        description="Other optional BLAST parameters, including user email.",
    )
    max_hits: Optional[int] = Field(
        default=15,
        description="Maximum number of hits to return in the BLAST results.",
    )
    sort_by: Optional[str] = Field(
        default="score",
        description="Criterion to sort BLAST results by, e.g., 'score', 'evalue'.",
    )
    megablast: Optional[str] = Field(
        default="on", description="Set to 'on' for human genome alignemnts"
    )
    question_uuid: Optional[str] = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="Unique identifier for the question.",
    )
    full_url: Optional[str] = Field(
        default="TBF",
        description="Full URL to be used to submit the BLAST query",
    )


class BlastQueryBuilder(BaseQueryBuilder):
    """A class for building a BlastQuery object."""

    def create_runnable(
        self,
        query_parameters: "BlastQueryParameters",
        conversation: "Conversation",
    ) -> Callable:
        """
        Creates a runnable object for executing queries using the LangChain
        `create_structured_output_runnable` method.

        Args:
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
            A Callable object that can execute the query.
        """
        return create_structured_output_runnable(
            output_schema=query_parameters,
            llm=conversation.chat,
            prompt=self.structured_output_prompt,
        )

    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> BlastQueryParameters:
        """
        Generates a BlastQuery object based on the given question, prompt, and
        BioChatter conversation. Uses a Pydantic model to define the API fields.
        Creates a runnable that can be invoked on LLMs that are qualified to
        parameterise functions.

        Args:
            question (str): The question to be answered.

            conversation: The conversation object used for parameterising the
                BlastQuery.

        Returns:
            BlastQuery: the parameterised query object (Pydantic model)
        """
        runnable = self.create_runnable(
            query_parameters=BlastQueryParameters,
            conversation=conversation,
        )
        blast_call_obj = runnable.invoke(
            {"input": f"Answer:\n{question} based on:\n {BLAST_QUERY_PROMPT}"}
        )
        blast_call_obj.question_uuid = str(uuid.uuid4())
        return blast_call_obj


class BlastFetcher(BaseFetcher):
    """
    A class for retrieving API results from BLAST given a parameterised
    BlastQuery.

    TODO add a limit of characters to be returned from the response.text?
    """

    def _submit_query(self, request_data: BlastQueryParameters) -> str:
        """Function to POST the BLAST query and retrieve RID.
        It submits the structured BlastQuery obj and return the RID.

        Args:
            request_data: BlastQuery object containing the BLAST query
                parameters.
        Returns:
            str: The Request ID (RID) for the submitted BLAST query.
        """
        data = {
            "CMD": request_data.cmd,
            "PROGRAM": request_data.program,
            "DATABASE": request_data.database,
            "QUERY": request_data.query,
            "FORMAT_TYPE": request_data.format_type,
            "MEGABLAST": request_data.megablast,
            "HITLIST_SIZE": request_data.max_hits,
        }
        # Include any other_params if provided
        if request_data.other_params:
            data.update(request_data.other_params)
        # Make the API call
        query_string = urlencode(data)
        # Combine base URL with the query string
        full_url = f"{request_data.url}?{query_string}"
        # Print the full URL
        request_data.full_url = full_url
        print("Full URL built by retriever:\n", request_data.full_url)
        response = requests.post(request_data.url, data=data)
        response.raise_for_status()
        # Extract RID from response
        print(response)
        match = re.search(r"RID = (\w+)", response.text)
        if match:
            return match.group(1)
        else:
            raise ValueError("RID not found in BLAST submission response.")

    def _fetch_results(
        self,
        rid: str,
        question_uuid: str,
        retries: int = 10000,
    ):
        """SECOND function to be called for a BLAST query.
        Will look for the RID to fetch the data
        """
        ###
        ###    TO DO: Implement logging for all BLAST queries
        ###
        # log_question_uuid_json(request_data.question_uuid,question, file_name, log_file_path,request_data.full_url)
        base_url = "https://blast.ncbi.nlm.nih.gov/Blast.cgi"
        check_status_params = {
            "CMD": "Get",
            "FORMAT_OBJECT": "SearchInfo",
            "RID": rid,
        }
        get_results_params = {
            "CMD": "Get",
            "FORMAT_TYPE": "XML",
            "RID": rid,
        }

        # Check the status of the BLAST job
        for attempt in range(retries):
            status_response = requests.get(base_url, params=check_status_params)
            status_response.raise_for_status()
            status_text = status_response.text
            print("evaluating status")
            if "Status=WAITING" in status_text:
                print(f"{question_uuid} results not ready, waiting...")
                time.sleep(15)
            elif "Status=FAILED" in status_text:
                raise RuntimeError("BLAST query FAILED.")
            elif "Status=UNKNOWN" in status_text:
                raise RuntimeError("BLAST query expired or does not exist.")
            elif "Status=READY" in status_text:
                if "ThereAreHits=yes" in status_text:
                    print(f"{question_uuid} results are ready, retrieving.")
                    results_response = requests.get(
                        base_url, params=get_results_params
                    )
                    results_response.raise_for_status()
                    # Save the results to a file
                    return results_response.text
                else:
                    return "No hits found"
        if attempt == retries - 1:
            raise TimeoutError(
                "Maximum attempts reached. Results may not be ready."
            )

    def fetch_results(
        self, query_model: BlastQueryParameters, retries: int = 20
    ) -> str:
        """
        Submit request and fetch results from BLAST API. Wraps individual
        submission and retrieval of results.

        Args:
            query_model: the Pydantic model of the query

            retries: the number of maximum retries

        Returns:
            str: the result from the BLAST API
        """
        rid = self._submit_query(request_data=query_model)
        return self._fetch_results(
            rid=rid,
            question_uuid=query_model.question_uuid,
            retries=retries,
        )


class BlastInterpreter(BaseInterpreter):
    def summarise_results(
        self,
        question: str,
        conversation_factory: Callable,
        response_text: str,
    ) -> str:
        """
        Function to extract the answer from the BLAST results.

        Args:
            question (str): The question to be answered.
            conversation_factory: A BioChatter conversation object.
            response_text (str): The response.text returned by NCBI.

        Returns:
            str: The extracted answer from the BLAST results.

        """
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a world class molecular biologist who knows everything about NCBI and BLAST results.",
                ),
                ("user", "{input}"),
            ]
        )
        summary_prompt = BLAST_SUMMARY_PROMPT.format(
            question=question, context=response_text
        )
        output_parser = StrOutputParser()
        conversation = conversation_factory()
        chain = prompt | conversation.chat | output_parser
        answer = chain.invoke({"input": {summary_prompt}})
        return answer



from abc import ABC, abstractmethod
from typing import Any, Optional
from collections.abc import Callable

from pydantic import BaseModel
from langchain_core.prompts import ChatPromptTemplate

from biochatter.llm_connect import Conversation


class BaseQueryBuilder(ABC):
    """
    An abstract base class for query builders.
    """

    @property
    def structured_output_prompt(self) -> ChatPromptTemplate:
        """
        Defines a structured output prompt template. This provides a default
        implementation for an API agent that can be overridden by subclasses to
        return a ChatPromptTemplate-compatible object.
        """
        return ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a world class algorithm for extracting information in structured formats.",
                ),
                (
                    "human",
                    "Use the given format to extract information from the following input: {input}",
                ),
                ("human", "Tip: Make sure to answer in the correct format"),
            ]
        )

    @abstractmethod
    def create_runnable(
        self,
        query_parameters: "BaseModel",
        conversation: "Conversation",
    ) -> Callable:
        """
        Creates a runnable object for executing queries. Must be implemented by
        subclasses. Should use the LangChain `create_structured_output_runnable`
        method to generate the Callable.

        Args:
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
            A Callable object that can execute the query.
        """
        pass

    @abstractmethod
    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> BaseModel:
        """

        Parameterises a query object (a Pydantic model with the fields of the
        API) based on the given question using a BioChatter conversation
        instance. Must be implemented by subclasses.

        Args:
            question (str): The question to be answered.

            conversation: The BioChatter conversation object containing the LLM
                that should parameterise the query.

        Returns:
            A parameterised instance of the query object (Pydantic BaseModel)
        """
        pass


class BaseFetcher(ABC):
    """
    Abstract base class for fetchers. A fetcher is responsible for submitting
    queries (in systems where submission and fetching are separate) and fetching
    and saving results of queries. It has to implement a `fetch_results()`
    method, which can wrap a multi-step procedure to submit and retrieve. Should
    implement retry method to account for connectivity issues or processing
    times.
    """

    @abstractmethod
    def fetch_results(
        self,
        query_model: BaseModel,
        retries: Optional[int] = 3,
    ):
        """
        Fetches results by submitting a query. Can implement a multi-step
        procedure if submitting and fetching are distinct processes (e.g., in
        the case of long processing times as in the case of BLAST).

        Args:
            query_model: the Pydantic model describing the parameterised query
        """
        pass


class BaseInterpreter(ABC):
    """
    Abstract base class for result interpreters. The interpreter is aware of the
    nature and structure of the results and can extract and summarise
    information from them.
    """

    @abstractmethod
    def summarise_results(
        self,
        question: str,
        conversation_factory: Callable,
        response_text: str,
    ) -> str:
        """
        Summarises an answer based on the given parameters.

        Args:
            question (str): The question that was asked.

            conversation_factory (Callable): A function that creates a
                BioChatter conversation.

            response_text (str): The response.text returned from the request.

        Returns:
            A summary of the answer.

        Todo:
            Genericise (remove file path and n_lines parameters, and use a
            generic way to get the results). The child classes should manage the
            specifics of the results.
        """
        pass



from .abc import BaseFetcher, BaseInterpreter, BaseQueryBuilder
from .blast import (
    BlastFetcher,
    BlastInterpreter,
    BlastQueryBuilder,
    BlastQueryParameters,
)
from .oncokb import OncoKBFetcher, OncoKBInterpreter, OncoKBQueryBuilder
from .bio_tools import (
    BioToolsFetcher,
    BioToolsInterpreter,
    BioToolsQueryBuilder,
)
from .api_agent import APIAgent

__all__ = [
    "BaseFetcher",
    "BaseInterpreter",
    "BaseQueryBuilder",
    "BlastFetcher",
    "BlastInterpreter",
    "BlastQueryBuilder",
    "BlastQueryParameters",
    "OncoKBFetcher",
    "OncoKBInterpreter",
    "OncoKBQueryBuilder",
    "BioToolsFetcher",
    "BioToolsInterpreter",
    "BioToolsQueryBuilder",
    "APIAgent",
]



from collections import deque
import os

from gtts import gTTS
from openai import OpenAI
from langchain.schema import Document
import nltk

from .llm_connect import GptConversation

FIRST_PROMPT = (
    "You are tasked with summarising a scientific manuscript for consumption as"
    "a podcast. As a first step, extract title and authors of the manuscript"
    "from the following text. Return them in the format 'Title: <title>,"
    "Authors: <authors>'."
)

SUMMARISE_PROMPT = (
    "You are tasked with processing a scientific manuscript for consumption as "
    "a podcast. You will receive a collection of sentences from the "
    "manuscript, from which you will remove any information not relevant to the "
    "content, such as references, figure legends, tables, author information, "
    "journal metadata, and so on. You will then be asked to summarise the "
    "section of the manuscript, making the wording more suitable for listening. "
    "Remove all content in brackets that is of technical nature, such as "
    "p-values, statistical tests, and so on. If the given text contains only "
    "literature references, return 'No content'."
)

PROCESS_PROMPT = (
    "You are tasked with processing a scientific manuscript for consumption as "
    "a podcast. You will receive a collection of sentences from the "
    "manuscript, from which you will remove any information not relevant to the "
    "content, such as references, figure legends, tables, author information, "
    "journal metadata, and so on. You will then be asked to process the "
    "section of the manuscript to make it listenable, but not shorten the content. "
    "Remove all content in brackets that is of technical nature, such as "
    "p-values, statistical tests, and so on. If the given text contains only "
    "literature references, return 'No content'."
)


class Podcaster:
    def __init__(
        self,
        document: Document,
        model_name: str = "gpt-3.5-turbo",
    ) -> None:
        """
        Orchestrates the podcasting of a document.
        """
        self.document = document
        self.model_name = model_name

    def generate_podcast(self, characters_per_paragraph: int) -> None:
        """
        Podcasts the document.

        TODO:
        - chain of density prompting for variable summary length
        """
        full_text = self.document[0].page_content

        # split text by sentence
        sentences = self._split_text(full_text)

        # could embed sentences and cluster on cosine similarity to identify
        # paragraphs here

        # preprocess text
        for i, sentence in enumerate(sentences):
            # special cases i.e. and e.g. - if sentence ends with one of these,
            # append next sentence
            special_cases = ["i.e.", "e.g."]
            if sentence.endswith(tuple(special_cases)):
                sentences[i] = sentence + " " + sentences[i + 1]
                del sentences[i + 1]

        # concatenate first 5 sentences for title and author extraction
        first_5 = "\n".join(sentences[:5])
        self.podcast_info = self._title_and_authors(first_5)

        # LLM to determine section breaks?

        # go through sections and summarise each
        self.processed_sections = self._process_sections(
            sentences,
            characters_per_paragraph,
        )

        # summarise the summaries

    def _split_text(self, text: str) -> list[str]:
        """
        Splits consecutive text into sentences.
        """
        nltk.download("punkt")
        tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")
        return tokenizer.tokenize(text)

    def _title_and_authors(self, text: str) -> str:
        """
        Extracts title and authors from document.

        Args:
            text (str): text to extract title and authors from

        Returns:
            str: title and authors
        """
        # first sentence - extract title, authors
        c_first = GptConversation(
            model_name=self.model_name,
            prompts={},
            correct=False,
        )
        c_first.set_api_key(api_key=os.getenv("OPENAI_API_KEY"), user="podcast")
        c_first.append_system_message(FIRST_PROMPT)
        msg, token_usage, correction = c_first.query(text)
        # split at authors ('Authors:' or '\nAuthors:')
        if "Authors:" in msg:
            title = msg.split("Title:")[1].split("Authors:")[0].strip()
            authors = msg.split("Authors:")[1].strip()
            return f"{title}, by {authors}, podcasted by biochatter."
        else:
            return "A podcast by biochatter."

    def _process_section(self, text: str, summarise: bool = False) -> str:
        """
        Processes a section of the document. Summarises if summarise is True,
        otherwise just makes the text more listenable.

        Args:
            text (str): text to summarise

            summarise (bool): whether to summarise the text

        Returns:
            str: summarised text
        """
        # summarise section
        c = GptConversation(
            model_name=self.model_name,
            prompts={},
            correct=False,
        )
        c.set_api_key(api_key=os.getenv("OPENAI_API_KEY"), user="podcast")
        if summarise:
            c.append_system_message(SUMMARISE_PROMPT)
        else:
            c.append_system_message(PROCESS_PROMPT)
        msg, token_usage, correction = c.query(text)
        return msg

    def _process_sections(
        self, sentences: list, characters_per_paragraph: int
    ) -> list:
        """

        Processes sections of the document. Concatenates sentences until
        characters_per_paragraph is reached, removing each sentence from the
        list as it is added to the section to be processed.

        Args:
            sentences (list): list of sentences to summarise

            characters_per_paragraph (int): number of characters per paragraph

        Returns:
            list: list of processed sections
        """
        summarised_sections = []
        section = ""
        while sentences:
            sentence = sentences.pop(0)
            tmp = section + sentence
            if len(tmp) < characters_per_paragraph and sentences:
                section += sentence
            else:
                if sentences:
                    sentences.insert(0, sentence)
                summarised_section = self._process_section(section)
                # filter "no content" sections
                if not (
                    "no content" in summarised_section.lower()
                    and len(summarised_section) < 30
                ):
                    summarised_sections.append(summarised_section)
                section = ""

        return summarised_sections

    def podcast_to_file(
        self,
        path: str,
        model: str = "gtts",
        voice: str = "alloy",
    ) -> None:
        """
        Uses text-to-speech to generate audio for the summarised paper podcast.

        Args:
            path (str): path to save audio file to

            model (str): model to use for text-to-speech. Currently supported:
                'gtts' (Google Text-to-Speech, free),
                'tts-1' (OpenAI API, paid, prioritises speed),
                'tts-1-hd' (OpenAI API, paid, prioritises quality)

            voice (str): voice to use for text-to-speech. See OpenAI API
                documentation for available voices.
        """

        full_text = self.podcast_to_text()

        if model == "gtts":
            audio = gTTS(text=full_text)
            audio.save(path)
        else:
            client = OpenAI()

            # Save the intro to the original file
            response = client.audio.speech.create(
                model=model,
                voice=voice,
                input=(
                    "You are listening to: \n\n"
                    + self.podcast_info
                    + "\n\n"
                    + " Text-to-speech generated by OpenAI."
                ),
            )
            first_path = path.rsplit(".", 1)[0] + "_0.mp3"
            response.stream_to_file(first_path)

            # Concatenate the sections
            full_text = ""
            for i, section in enumerate(self.processed_sections):
                full_text += section + "\n\n"

            # Make sections of 4000 characters max (at sentence boundaries)
            nltk.download("punkt")
            tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")
            sentences = deque(
                tokenizer.tokenize(full_text)
            )  # Use a deque instead of a list

            # Split the text into sections by filling one section until it
            # exceeds 4000 characters, then starting a new section (not adding
            # the sentence that would exceed the limit)
            sections = []
            section = ""
            while sentences:
                sentence = sentences[0]
                tmp = section + sentence
                if len(tmp) < 4000:
                    section += sentences.popleft()
                else:
                    sections.append(section)
                    section = ""

            sections.append(section)  # Add the penultimate section

            # Last section: conclude the podcast
            sections.append(
                f"This was {self.podcast_info}. Thank you for listening."
            )

            # Save each section to a separate file with an integer suffix
            for i, section in enumerate(sections):
                response = client.audio.speech.create(
                    model=model,
                    voice=voice,
                    input=section,
                )
                # Insert the integer suffix just before the .mp3 extension
                section_path = path.rsplit(".", 1)[0] + f"_{i+1}.mp3"
                response.stream_to_file(section_path)

    def podcast_to_text(self):
        """
        Returns the summarised paper podcast as text.
        """
        full_text = "You are listening to: " + self.podcast_info + "\n\n"
        for section in self.processed_sections:
            full_text += section + "\n\n"
        return full_text



# straight copy from BioCypher repo
# TODO: have a utils package for both repos

from typing import (
    Any,
    Union,
    Mapping,
    KeysView,
    Generator,
    ItemsView,
    ValuesView,
)
from collections.abc import Iterable
import re

import stringcase

__all__ = ["LIST_LIKE", "SIMPLE_TYPES", "ensure_iterable", "to_list"]

SIMPLE_TYPES = (
    bytes,
    str,
    int,
    float,
    bool,
    type(None),
)

LIST_LIKE = (
    list,
    set,
    tuple,
    Generator,
    ItemsView,
    KeysView,
    Mapping,
    ValuesView,
)


def to_list(value: Any) -> list:
    """
    Ensures that ``value`` is a list.
    """

    if isinstance(value, LIST_LIKE):
        value = list(value)

    else:
        value = [value]

    return value


def ensure_iterable(value: Any) -> Iterable:
    """
    Returns iterables, except strings, wraps simple types into tuple.
    """

    return value if isinstance(value, LIST_LIKE) else (value,)


# string conversion, adapted from Biolink Model Toolkit
lowercase_pattern = re.compile(r"[a-zA-Z]*[a-z][a-zA-Z]*")
underscore_pattern = re.compile(r"(?<!^)(?=[A-Z][a-z])")


def from_pascal(s: str, sep: str = " ") -> str:
    underscored = underscore_pattern.sub(sep, s)
    lowercased = lowercase_pattern.sub(
        lambda match: match.group(0).lower(),
        underscored,
    )
    return lowercased


def pascalcase_to_sentencecase(s: str) -> str:
    """
    Convert PascalCase to sentence case.

    Args:
        s: Input string in PascalCase

    Returns:
        string in sentence case form
    """
    return from_pascal(s, sep=" ")


def snakecase_to_sentencecase(s: str) -> str:
    """
    Convert snake_case to sentence case.

    Args:
        s: Input string in snake_case

    Returns:
        string in sentence case form
    """
    return stringcase.sentencecase(s).lower()


def sentencecase_to_snakecase(s: str) -> str:
    """
    Convert sentence case to snake_case.

    Args:
        s: Input string in sentence case

    Returns:
        string in snake_case form
    """
    return stringcase.snakecase(s).lower()


def sentencecase_to_pascalcase(s: str) -> str:
    """
    Convert sentence case to PascalCase.

    Args:
        s: Input string in sentence case

    Returns:
        string in PascalCase form
    """
    return re.sub(r"(?:^| )([a-zA-Z])", lambda match: match.group(1).upper(), s)


def to_lower_sentence_case(s: str) -> str:
    """
    Convert any string to lower sentence case. Works with snake_case,
    PascalCase, and sentence case.

    Args:
        s: Input string

    Returns:
        string in lower sentence case form
    """
    if "_" in s:
        return snakecase_to_sentencecase(s)
    elif " " in s:
        return s.lower()
    elif s[0].isupper():
        return pascalcase_to_sentencecase(s)
    else:
        return s
